/**************************************************************************
 * This file is part of ESTmapper.
 * Copyright (c) 2001, 2002, 2003, 2004 Applera Corporation
 * Authors: Brian Walenz and Liliana Florea
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received (LICENSE.txt) a copy of the GNU General Public 
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************************************************/


#ifndef BRI_BITPACKING_H
#define BRI_BITPACKING_H

//  Routines used for stuffing bits into a word array.

//  Define this to enable testing that the width of the data element
//  is greater than zero.  The u64bitMASK() macro (bri.h) does not
//  generate a mask for 0.  Compiler warnings are issued, because you
//  shouldn't use this in production code.
//
//#define CHECK_WIDTH


//  Returns 'siz' bits from the stream based at 'ptr' and currently at
//  location 'pos'.  The position of the stream is not changed.
//
u64bit
getDecodedValue(u64bit *ptr,
                u64bit  pos,
                u64bit  siz);


//  Copies the lowest 'siz' bits in 'val' to the stream based at 'ptr'
//  and currently at 'pos'.  The position of the stream is not
//  changed.
//
void
setDecodedValue(u64bit *ptr,
                u64bit  pos,
                u64bit  siz,
                u64bit  val);


//  Like getDecodedValue() but will pre/post increment/decrement the
//  value stored in the stream before in addition to returning the
//  value.
//
//  preIncrementDecodedValue(ptr, pos, siz) === x = getDecodedValue(ptr, pos, siz) + 1;
//                                              setDecodedValue(ptr, pos, siz, x);
//
//  preDecrementDecodedValue(ptr, pos, siz) === x = getDecodedValue(ptr, pos, siz) - 1;
//                                              setDecodedValue(ptr, pos, siz, x);
//
//  postIncrementDecodedValue(ptr, pos, siz) === x = getDecodedValue(ptr, pos, siz);
//                                               setDecodedValue(ptr, pos, siz, x + 1);
//
//  postDecrementDecodedValue(ptr, pos, siz) === x = getDecodedValue(ptr, pos, siz);
//                                               setDecodedValue(ptr, pos, siz, x - 1);
//
u64bit
preIncrementDecodedValue(u64bit *ptr,
                         u64bit  pos,
                         u64bit  siz);

u64bit
preDecrementDecodedValue(u64bit *ptr,
                         u64bit  pos,
                         u64bit  siz);

u64bit
postIncrementDecodedValue(u64bit *ptr,
                          u64bit  pos,
                          u64bit  siz);

u64bit
postDecrementDecodedValue(u64bit *ptr,
                          u64bit  pos,
                          u64bit  siz);





//  NOTE:
//
//  I assume the addresses of bits in two consectuve words would be:
//
//    [ 3 2 1 0 ][ 3 2 1 0 ]
//
//  However, setDecodedValue() works as:
//
//    bitstream = [xxxx][xx00][0000]
//    value = 0cba (three bits have information. the fourth is empty)
//
//    setDecodedValue(bitstream, value, 3);
//
//    bitstream = [xxxx][xxcb][a000]
//
//  So, if you're reading the bitstream bit-by-bit, you get the values
//  high-order first.
//



inline
u64bit
getDecodedValue(u64bit *ptr,
                u64bit  pos,
                u64bit  siz) {
  u64bit wrd = (pos >> 6) & 0x0000cfffffffffffllu;
  u64bit bit = (pos     ) & 0x000000000000003fllu;
  u64bit b1  = 64 - bit;
  u64bit b2  = siz - b1;  //  Only used if siz > b1
  u64bit ret = 0;

#ifdef CHECK_WIDTH
#warning libbri/bitPacking.h::getDecodedValue() has defined CHECK_WIDTH!
  if (siz == 0) {
    fprintf(stderr, "ERROR: getDecodedValue() called with zero size!\n");
    abort();
  }
#endif

  if (b1 >= siz) {
    ret = ptr[wrd] >> (b1 - siz);
  } else {
    ret  = (ptr[wrd] & u64bitMASK(b1)) << b2;
    ret |= (ptr[wrd+1] >> (64 - b2)) & u64bitMASK(b2);
  }

  ret &= u64bitMASK(siz);

  return(ret);
}



inline
void
setDecodedValue(u64bit *ptr,
                u64bit  pos,
                u64bit  siz,
                u64bit  val) {
  u64bit wrd = (pos >> 6) & 0x0000cfffffffffffllu;
  u64bit bit = (pos     ) & 0x000000000000003fllu;
  u64bit b1  = 64 - bit;
  u64bit b2  = siz - b1;  //  Only used if siz > b1

#ifdef CHECK_WIDTH
#warning libbri/bitPacking.h::setDecodedValue() has defined CHECK_WIDTH!
  if (siz == 0) {
    fprintf(stderr, "ERROR: setDecodedValue() called with zero size!\n");
    abort();
  }
#endif

  val &= u64bitMASK(siz);

  if (b1 >= siz) {
    ptr[wrd] &= ~( u64bitMASK(siz) << (b1-siz) );
    ptr[wrd] |= val << (b1-siz);
  } else {
    ptr[wrd] &= ~u64bitMASK(b1);
    ptr[wrd] |= (val & (u64bitMASK(b1) << (b2))) >> (b2);

    ptr[wrd+1] &= ~(u64bitMASK(b2) << (64-b2));
    ptr[wrd+1] |= (val & (u64bitMASK(b2))) << (64-b2);
  }
}



inline
u64bit
preIncrementDecodedValue(u64bit *ptr,
                         u64bit  pos,
                         u64bit  siz) {
  u64bit wrd = (pos >> 6) & 0x0000cfffffffffffllu;
  u64bit bit = (pos     ) & 0x000000000000003fllu;
  u64bit b1  = 64 - bit;
  u64bit b2  = siz - b1;  //  Only used if siz > b1
  u64bit ret  = 0;

#ifdef CHECK_WIDTH
#warning libbri/bitPacking.h::preIncrementDecodedValue() has defined CHECK_WIDTH!
  if (siz == 0) {
    fprintf(stderr, "ERROR: preIncrementDecodedValue() called with zero size!\n");
    abort();
  }
#endif

  if (b1 >= siz) {
    ret  = ptr[wrd] >> (b1 - siz);

    ret++;
    ret &= u64bitMASK(siz);

    ptr[wrd] &= ~( u64bitMASK(siz) << (b1-siz) );
    ptr[wrd] |= ret << (b1-siz);
  } else {
    ret  = (ptr[wrd] & u64bitMASK(b1)) << b2;
    ret |= (ptr[wrd+1] >> (64 - b2)) & u64bitMASK(b2);

    ret++;
    ret &= u64bitMASK(siz);

    ptr[wrd] &= ~u64bitMASK(b1);
    ptr[wrd] |= (ret & (u64bitMASK(b1) << (b2))) >> (b2);

    ptr[wrd+1] &= ~(u64bitMASK(b2) << (64-b2));
    ptr[wrd+1] |= (ret & (u64bitMASK(b2))) << (64-b2);
  }

  return(ret);
}



inline
u64bit
preDecrementDecodedValue(u64bit *ptr,
                         u64bit  pos,
                         u64bit  siz) {
  u64bit wrd = (pos >> 6) & 0x0000cfffffffffffllu;
  u64bit bit = (pos     ) & 0x000000000000003fllu;
  u64bit b1  = 64 - bit;
  u64bit b2  = siz - b1;  //  Only used if siz > b1
  u64bit ret = 0;

#ifdef CHECK_WIDTH
#warning libbri/bitPacking.h::preDecrementDecodedValue() has defined CHECK_WIDTH!
  if (siz == 0) {
    fprintf(stderr, "ERROR: preDecrementDecodedValue() called with zero size!\n");
    abort();
  }
#endif

  if (b1 >= siz) {
    ret = ptr[wrd] >> (b1 - siz);

    ret--;
    ret &= u64bitMASK(siz);

    ptr[wrd] &= ~( u64bitMASK(siz) << (b1-siz) );
    ptr[wrd] |= ret << (b1-siz);
  } else {
    ret  = (ptr[wrd] & u64bitMASK(b1)) << b2;
    ret |= (ptr[wrd+1] >> (64 - b2)) & u64bitMASK(b2);

    ret--;
    ret &= u64bitMASK(siz);

    ptr[wrd] &= ~u64bitMASK(b1);
    ptr[wrd] |= (ret & (u64bitMASK(b1) << (b2))) >> (b2);

    ptr[wrd+1] &= ~(u64bitMASK(b2) << (64-b2));
    ptr[wrd+1] |= (ret & (u64bitMASK(b2))) << (64-b2);
  }

  return(ret);
}



inline
u64bit
postIncrementDecodedValue(u64bit *ptr,
                          u64bit  pos,
                          u64bit  siz) {
  u64bit wrd = (pos >> 6) & 0x0000cfffffffffffllu;
  u64bit bit = (pos     ) & 0x000000000000003fllu;
  u64bit b1  = 64 - bit;
  u64bit b2  = siz - b1;  //  Only used if siz > b1
  u64bit ret = 0;

#ifdef CHECK_WIDTH
#warning libbri/bitPacking.h::postIncrementDecodedValue() has defined CHECK_WIDTH!
  if (siz == 0) {
    fprintf(stderr, "ERROR: postIncrementDecodedValue() called with zero size!\n");
    abort();
  }
#endif

  if (b1 >= siz) {
    ret = ptr[wrd] >> (b1 - siz);

    ret++;
    ret &= u64bitMASK(siz);

    ptr[wrd] &= ~( u64bitMASK(siz) << (b1-siz) );
    ptr[wrd] |= ret << (b1-siz);
  } else {
    ret  = (ptr[wrd] & u64bitMASK(b1)) << b2;
    ret |= (ptr[wrd+1] >> (64 - b2)) & u64bitMASK(b2);

    ret++;
    ret &= u64bitMASK(siz);

    ptr[wrd] &= ~u64bitMASK(b1);
    ptr[wrd] |= (ret & (u64bitMASK(b1) << (b2))) >> (b2);

    ptr[wrd+1] &= ~(u64bitMASK(b2) << (64-b2));
    ptr[wrd+1] |= (ret & (u64bitMASK(b2))) << (64-b2);
  }

  ret--;
  ret &= u64bitMASK(siz);

  return(ret);
}





inline
u64bit
postDecrementDecodedValue(u64bit *ptr,
                          u64bit  pos,
                          u64bit  siz) {
  u64bit wrd = (pos >> 6) & 0x0000cfffffffffffllu;
  u64bit bit = (pos     ) & 0x000000000000003fllu;
  u64bit b1  = 64 - bit;
  u64bit b2  = siz - b1;  //  Only used if siz > b1
  u64bit ret = 0;

#ifdef CHECK_WIDTH
#warning libbri/bitPacking.h::postDecrementDecodedValue() has defined CHECK_WIDTH!
  if (siz == 0) {
    fprintf(stderr, "ERROR: postDecrementDecodedValue() called with zero size!\n");
    abort();
  }
#endif

  if (b1 >= siz) {
    ret = ptr[wrd] >> (b1 - siz);

    ret--;
    ret &= u64bitMASK(siz);

    ptr[wrd] &= ~( u64bitMASK(siz) << (b1-siz) );
    ptr[wrd] |= ret << (b1-siz);
  } else {
    ret  = (ptr[wrd] & u64bitMASK(b1)) << b2;
    ret |= (ptr[wrd+1] >> (64 - b2)) & u64bitMASK(b2);

    ret--;
    ret &= u64bitMASK(siz);

    ptr[wrd] &= ~u64bitMASK(b1);
    ptr[wrd] |= (ret & (u64bitMASK(b1) << (b2))) >> (b2);

    ptr[wrd+1] &= ~(u64bitMASK(b2) << (64-b2));
    ptr[wrd+1] |= (ret & (u64bitMASK(b2))) << (64-b2);
  }

  ret++;
  ret &= u64bitMASK(siz);

  return(ret);
}



#endif  //  BRI_BITPACKING_H
