// This file is part of ESTmapper.
// Copyright (c) 2001, 2002, 2003, 2004 Applera Corporation
// Authors: Brian Walenz and Liliana Florea
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received (LICENSE.txt) a copy of the GNU General Public 
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "positionDB.H"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

static
char     magic[16] = { 'p', 'o', 's', 'i', 't', 'i', 'o', 'n', 'D', 'B', '1', ' ', ' ', ' ', ' ', ' '  };
static
char     faild[16] = { 'p', 'o', 's', 'i', 't', 'i', 'o', 'n', 'D', 'B', 'f', 'a', 'i', 'l', 'e', 'd'  };


void
positionDB::saveState(char const *filename) {

  errno = 0;
  int F = open(filename, O_RDWR | O_CREAT | O_LARGEFILE,
               S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
  if (errno) {
    fprintf(stderr, "Can't open '%s' for writing positionDB.\n%s\n", filename, strerror(errno));
    exit(1);
  }

  bool    magicFirst = false;

  //  Test if this is a pipe.  If so, we write the magic first,
  //  otherwise we write the magic last.
  //
  errno      = 0;
  lseek(F, 0, SEEK_SET);
  if (errno == ESPIPE)
    magicFirst = true;

  if (magicFirst)
    write(F, magic, sizeof(char) * 16);
  else
    write(F, faild, sizeof(char) * 16);

  write(F, this,       sizeof(positionDB) * 1);
  write(F, _hashTable, sizeof(u64bit) * (_tableSizeInEntries * _hashWidth / 64 + 1));
  write(F, _buckets,   sizeof(u64bit) * (_numberOfDistinct   * _wFin      / 64 + 1));
  write(F, _positions, sizeof(u64bit) * (_numberOfEntries    * _posnWidth / 64 + 1));

  if (errno) {
    fprintf(stderr, "positionDB::saveState()-- Write failure.\n%s\n", strerror(errno));
    exit(1);
  }

  if (magicFirst == false) {
    lseek(F, 0, SEEK_SET);
    if (errno) {
      fprintf(stderr, "positionDB::saveState()-- Failed to seek to start of file -- write failed.\n%s\n", strerror(errno));
      exit(1);
    }
    write(F, magic, sizeof(char) * 16);
  }

  close(F);

  if (errno) {
    fprintf(stderr, "positionDB::saveState()-- Write failure.\n%s\n", strerror(errno));
    exit(1);
  }
}


bool
positionDB::loadState(char const *filename, bool beNoisy, bool loadData) {
  char   cigam[16] = { 0 };

  errno = 0;
  int F = open(filename, O_RDONLY | O_LARGEFILE, 0);
  if (errno) {
    fprintf(stderr, "Can't open '%s' for reading pre-built positionDB.\n%s\n", filename, strerror(errno));
    return(false);
  }

  read(F, cigam, sizeof(char) * 16);

  if        (strncmp(faild, cigam, 16) == 0) {
    if (beNoisy)
      fprintf(stderr, "positionDB::loadState()-- Incomplete positionDB binary file.\n");
    close(F);
    return(false);
  } else if (strncmp(magic, cigam, 16) != 0) {
    if (beNoisy) {
      fprintf(stderr, "positionDB::loadState()-- Not a positionDB binary file, maybe a sequence file?\n");
      fprintf(stderr, "positionDB::loadState()-- Read     '%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c'\n",
              cigam[0],  cigam[1],  cigam[2],  cigam[3],
              cigam[4],  cigam[5],  cigam[6],  cigam[7],
              cigam[8],  cigam[9],  cigam[10], cigam[11],
              cigam[12], cigam[13], cigam[14], cigam[15]);
      fprintf(stderr, "positionDB::loadState()-- Expected '%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c'\n",
              magic[0],  magic[1],  magic[2],  magic[3],
              magic[4],  magic[5],  magic[6],  magic[7],
              magic[8],  magic[9],  magic[10], magic[11],
              magic[12], magic[13], magic[14], magic[15]);
    }

    close(F);
    return(false);
  }

  read(F, this, sizeof(positionDB) * 1);

  _hashTable = 0L;
  _buckets   = 0L;
  _positions = 0L;

  if (loadData) {
    u64bit  hs = _tableSizeInEntries * _hashWidth / 64 + 1;
    u64bit  bs = _numberOfDistinct   * _wFin      / 64 + 1;
    u64bit  ps = _numberOfEntries    * _posnWidth / 64 + 1;

    _hashTable = new u64bit [hs];
    _buckets   = new u64bit [bs];
    _positions = new u64bit [ps];

    read(F, _hashTable, sizeof(u64bit) * hs);
    read(F, _buckets,   sizeof(u64bit) * bs);
    read(F, _positions, sizeof(u64bit) * ps);
  }

  close(F);

  if (errno) {
    fprintf(stderr, "positionDB::loadState()-- Read failure.\n%s\n", strerror(errno));
    exit(1);
  }

  return(true);
}



void
positionDB::printState(FILE *stream) {
  fprintf(stream, "merSizeInBases:       "u32bitFMT"\n",  _merSizeInBases);
  fprintf(stream, "merSkipInBases:       "u32bitFMT"\n",  _merSkipInBases);
  fprintf(stream, "tableSizeInBits:      "u32bitFMT"\n",  _tableSizeInBits);
  fprintf(stream, "tableSizeInEntries:   "u64bitFMT"\n", _tableSizeInEntries);
  fprintf(stream, "hashWidth:            "u32bitFMT"\n",  _hashWidth);
  fprintf(stream, "chckWidth:            "u32bitFMT"\n",  _chckWidth);
  fprintf(stream, "posnWidth:            "u32bitFMT"\n",  _posnWidth);
  fprintf(stream, "numberOfMers:         "u64bitFMT"\n", _numberOfMers);
  fprintf(stream, "numberOfPositions:    "u64bitFMT"\n", _numberOfPositions);
  fprintf(stream, "numberOfDistinct:     "u64bitFMT"\n", _numberOfDistinct);
  fprintf(stream, "numberOfUnique:       "u64bitFMT"\n", _numberOfUnique);
  fprintf(stream, "numberOfEntries:      "u64bitFMT"\n", _numberOfEntries);
  fprintf(stream, "maximumEntries:       "u64bitFMT"\n", _maximumEntries);
}

