#!/usr/bin/perl

# This file is part of ESTmapper.
# Copyright (c) 2001, 2002, 2003, 2004 Applera Corporation
# Authors: Brian Walenz and Liliana Florea
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received (LICENSE.txt) a copy of the GNU General Public 
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



use lib "/work/assembly/projects/scripts";
use libBri;
use strict;

#
#  Example code of how to use the libBri sim4reader routine.
#  Describes the various fields.  A much better (faster and more
#  extensive) reader is available in C in cds/IR/BRI/sim4polish/
#
#  This code will read in a set of polishes from STDIN and compute the
#  average match location on each genomic sequence.  Completely
#  useless, but a good example.
#
#
#  The match record contains a description of the query
#  sequence (est, mrna, fragment) and of the genomic region
#  the query is aligned to.  The 'raw' entry is defined only
#  if the record was read in with no errors.
#
#  The full record in an unparsed format is contained in
#  the 'raw' entry.  It can be printed to an ASCII terminal,
#  and doesn't need a newline.
#
#  The query is described by:
#
#    'estID'      - IID of the query.  These start at zero and
#                   are usually (should be) static if the file
#                   the queries come from doesn't change.
#    'estDefLine' - defline of the query, including '>'.
#    'estLen'     - length in bp of the query.
#    'pA'         - length of the poly-A tail (at the end of the
#                   seq).
#    'pT'         - length of the poly-T tail (at the start of
#                   the seq).
#
#  The genomic is described by:
#
#    'dbID'       - IID of the genomic sequence.  Same properties
#                   as for estID.
#    'dbDefLine'  - defline of the sequence, including '>'.
#    'dbLen'      - length in bp of the genomic sequence.
#    'dbLo'       - start of the region examined (space-based).
#    'dbHi'       - end of the region examined (space-based).
#
#  The quality and analysis of the match are described by:
#
#    'numMatches'        - number of non-N letters that match in
#                          the alignment.
#    'numMatchesN'       - number of N's that match in the alignment.
#    'percentID'         - percent identity of the alignment.
#    'coverage'          - amount of the query that is covered by
#                          the alignment.
#    'matchOrientation'  - orientation of the alignment.
#    'strandOrientation' - predicted strand the query matched.
#
#  Additionally the match is described by:
#
#    'comment'           - free form text, with no newlines.
#
#  The alignment itself is stored in an array of 'exons'.  Each exon is
#  described by:
#
#    'cDNAstart'         - the base-based start of this alignment in
#                          the query.
#    'cDNAend'           - the base-based end of this alignment in the
#                          query.
#    'GENOMICstart'      - the base-based start of this alignment in the
#                          genomic _region_.  To get the position in the
#                          genomic _sequence_, add 'dbLo' to this value.
#    'GENOMICend'        - the base-based start of this alignment in the
#                          genomic _region_.  To get the position in the
#                          genomic _sequence_, add 'dbLo' to this value.
#    'numMatches'        - number of non-N letters that match in the
#                          alignment.
#    'numMatchesN'       - number of N's that match in the alignment
#    'percentID'         - percent identity of the alignemnt.
#    'intronOrientation' - predicted orientation of the intron, same
#                          as sim4.
#


my @aveSta;
my @aveEnd;
my @aveNum;
my @seqLen;

while (!eof(STDIN)) {

    #  Read the next match into the hash %p.  Note the * before the
    #  file handle.
    #
    my %p = &libBri::readPolish(*STDIN);

    #  If 'raw' is undefined, then there was an error in reading the
    #  match.
    #
    if (defined($p{'raw'})) {

        #  For each exon, sum the length of the alignment.
        #
        foreach my $e (@{$p{'exons'}}) {
            $aveSta[$p{'dbID'}] += $p{'dbLo'} + $e->{'GENOMICSta'};
            $aveEnd[$p{'dbID'}] += $p{'dbLo'} + $e->{'GENOMICend'};
            $aveNum[$p{'dbID'}]++;
        }
    }
}

for (my $i=0; $i < scalar(@aveNum); $i++) {
    if (defined($aveNum[$i])) {
        my $as = int($aveSta[$i] / $aveNum[$i]);
        my $ae = int($aveEnd[$i] / $aveNum[$i]);

        print STDERR "$i\t$as\t$ae\n";
    } else {
        print STDERR "$i\t0\t0\n";
    }
}
