// This file is part of ESTmapper.
// Copyright (c) 2001, 2002, 2003, 2004 Applera Corporation
// Authors: Brian Walenz and Liliana Florea
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received (LICENSE.txt) a copy of the GNU General Public 
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "bri++.H"
#include "sim4polish.h"
#include "sim4polishBuilder.H"
#include "sim4polishFile.H"

//  Remove redundant polishes from an input set.
//
//  Redundancy is defined as two polishes that overlap on the genome.
//  Any amount of overlap is redundant.
//
//  The longest of the overlapping matches is saved.

//#define DEBUGOUT

u32bit  findOverlap(sim4polish *A, sim4polish *B);

int
main(int argc, char **argv) {

  if (argc != 2) {
    fprintf(stderr, "usage: %s <polishes-file>\n", argv[0]);
    fprintf(stderr, "(yes, you _must_ give it a file.  stdin is not possible.)\n");
    exit(1);
  }

  u32bit  matchesWithNoOverlap       = 0;
  u32bit  matchesWithOverlap         = 0;
  u32bit  notPerfectClique           = 0;

  //  Open a polishFile and force the index to build
  //
  sim4polishFile *Afile = new sim4polishFile(argv[1]);
  Afile->setPosition(0);

  //  Ask both for the largest EST iid seen, then iterate over those.
  //
  u32bit  largestIID = Afile->maxIID();

  for (u32bit iid=0; iid<largestIID; iid++) {
    sim4polishList *A = Afile->getEST(iid);

    if (A->length() > 0) {

      //  fill out the overlap matrix

      u32bit  **overlap = new u32bit* [A->length()];
      overlap[0] = new u32bit [A->length() * A->length()];
      for (u32bit i=1; i<A->length(); i++)
        overlap[i] = overlap[i-1] + A->length();

      for (u32bit a=0; a<A->length(); a++)
        for (u32bit b=0; b<A->length(); b++)
          if (a == b)
            overlap[a][b] = 0;
          else
            overlap[a][b] = findOverlap((*A)[a], (*A)[b]);

      //  look for guys with no overlaps, print and remove them

      sim4polishList *W = new sim4polishList;

      for (u32bit a=0; a<A->length(); a++) {
        bool nooverlaps = true;

        for (u32bit b=0; b<A->length(); b++)
          if (overlap[a][b])
            nooverlaps = false;

        if (nooverlaps) {
          matchesWithNoOverlap++;

          s4p_printPolish(stdout, (*A)[a], 0);
        } else {
          matchesWithOverlap++;
          W->push(s4p_copyPolish((*A)[a]));
        }
      }


#if 1
      fprintf(stderr, "IID="u32bitFMTW(8)" -- overlap:"u32bitFMT" noOverlap:"u32bitFMT"\r",
              iid, matchesWithOverlap, matchesWithNoOverlap);
      fflush(stderr);
#endif


      //  A is junk, W contains the matches that overlap.

      delete A;
      A = 0L;


      //  Report all the overlaps

#ifdef DEBUGOUT
      for (u32bit a=0; a<W->length(); a++) {
        sim4polish *p = (*W)[a];
        fprintf(stderr, u32bitFMTW(3)": "u32bitFMTW(3)"--"u32bitFMTW(3)"\n",
                iid, p->exons[0].genFrom, p->exons[p->numExons-1].genTo);
      }
#endif



      //  while we have matches in the set of overlapping matches,
      //  find a connected component, check that it is/is not a
      //  clique, and decide which match to keep.

      u32bit  *clique      = new u32bit [W->length()];
      u32bit   cliqueSize  = 0;
      bool     inserted    = false;
      u32bit   *length     = new u32bit [W->length()];

      while (W->length() > 0) {

#ifdef DEBUGOUT
        fprintf(stderr, "IID="u32bitFMTW(8)" -- examine "u32bitFMT" matches\n",
                iid, W->length());
#endif

        //  Find the length of all the matches in this set

        for (u32bit a=0; a<W->length(); a++) {
          length[a] = 0;
          for (u32bit i=0; i<(*W)[a]->numExons; i++)
            length[a] += (*W)[a]->exons[i].genTo - (*W)[a]->exons[i].genFrom + 1;
        }

        //  reconstruct the overlap matrix -- hey, if you want to be
        //  efficient and recover this from the existing one, nobody is
        //  stopping you.
        
        for (u32bit a=0; a<W->length(); a++)
          for (u32bit b=0; b<W->length(); b++)
            if (a == b)
              overlap[a][b] = 0;
            else
              overlap[a][b] = findOverlap((*W)[a], (*W)[b]);

        //  OK, now find the clique/connected component

        for (u32bit i=0; i<W->length(); i++)
          clique[i] = 0;

        clique[0]   = 1;
        cliqueSize  = 1;
        inserted    = true;

        while (inserted) {
          inserted = false;

          //  If a is in the clique, add all it's overlaps

          for (u32bit a=0; a<W->length(); a++) {
            if (clique[a]) {
              for (u32bit b=0; b<W->length(); b++) {
                if ((overlap[a][b]) && (!clique[b])) {
                  clique[b] = 1;
                  cliqueSize++;
                  inserted  = true;
                }
              }
            }
          }
        }

#ifdef DEBUGOUT
        fprintf(stderr, "IID="u32bitFMTW(8)" -- examine "u32bitFMT" matches, found "u32bitFMT" overlapping\n",
                iid, W->length(), cliqueSize);
#endif

        //  Check that it is a clique

        if (cliqueSize > 2) {

          u32bit num = 0;

          for (u32bit a=0; a<W->length(); a++)
            for (u32bit b=0; b<W->length(); b++)
              if (clique[a] && clique[b] && overlap[a][b])
                num++;

          if (num != cliqueSize * (cliqueSize-1)) {
            notPerfectClique++;

            fprintf(stderr, "\nNOT A PERFECT CLIQUE!  Found "u32bitFMT" overlaps, wanted "u32bitFMT" in the clique.\n",
                    num, cliqueSize * (cliqueSize-1));

            for (u32bit a=0; a<W->length(); a++)
              if (clique[a])
                s4p_printPolish(stderr, (*W)[a], S4P_PRINTPOLISH_MINIMAL);
          }
          
        }

        //  Find the longest member, output it

        u32bit longest = 0;
        while (clique[longest] == 0)
          longest++;

        for (u32bit i=0; i<W->length(); i++)
          if ((clique[i]) && (length[longest] < length[i]))
            longest = i;

        s4p_printPolish(stdout, (*W)[longest], 0);

        //  Remove the clique from the set of overlaps

        A = new sim4polishList;
        for (u32bit i=0; i<W->length(); i++) {
          if (clique[i] == 0)
            A->push(s4p_copyPolish( (*W)[i] ));
        }

        delete W;
        W = A;
        A = 0L;
      }

      delete [] clique;
      delete    W;

      delete [] overlap[0];
      delete [] overlap;
    }

    delete A;
  }

  fprintf(stderr, "\nmatches withOvl:"u32bitFMT" withoutOvl:"u32bitFMT"\n",
          matchesWithOverlap, matchesWithNoOverlap);
  fprintf(stderr, "not perfect clique:"u32bitFMT"\n", notPerfectClique);

  delete Afile;
}


