/**************************************************************************
 * This file is part of ESTmapper.
 * Copyright (c) 2001, 2002, 2003, 2004 Applera Corporation
 * Authors: Brian Walenz and Liliana Florea
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received (LICENSE.txt) a copy of the GNU General Public 
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <errno.h>


pid_t
runChild(int useShell, char *cmd, int cmdlen) {
  pid_t   child;
  char   *args[4096];
  int     i, arg;

  child = fork();

  if (child != 0) {
    return(child);
  } else {

    //  First, strip off any whitespace from the end of the string.
    //
    i = cmdlen-1;
    while ((i>=0) && isspace(cmd[i])) {
      cmd[i] = 0;
      i--;
    }

    switch (useShell) {
    case 0:
      //  no shell, just exec
      //

      //  The command to run is the first word.
      //
      args[0] = cmd;
      arg = 1;

      //  Place the remaining words into the array, changing spaces into
      //  zero's.
      //
      for (i=0; i<cmdlen; i++) {
        
        //  If i is a space, we have found the start of a new word.
        //
        if (isspace(cmd[i])) {

          //  Change all spaces to zeros
          //
          while (isspace(cmd[i])) {
            cmd[i] = 0;
            i++;
          }

          //  i is now the start of the next word.  Remember that
          //  the cmd string ends with zero, not space.
          //
          args[arg] = cmd + i;
          arg++;
        }
      }

      //  Terminate the list of args
      //
      args[arg] = 0L;
      break;
    case 1:
      //  /bin/sh
      //
      args[0] = "/bin/sh";
      args[1] = "-c";
      args[2] = cmd;
      args[3] = 0L;
      break;
    }

    //  Run it.
    //
    errno = 0;
    execv(args[0], args);

    if (errno) {
      fprintf(stderr, "Couldn't run '%s'", args[0]);
      perror(0L);
    }

    exit(0);
  }
}


int
reapChildren(pid_t *kids, int numCPUs) {
  int i, status, avail = 0;

  for (i=0; i<numCPUs; i++) {
    if ((kids[i] != -1) &&
        (waitpid(kids[i], &status, WNOHANG) > 0) &&
        (WIFEXITED(status)))
      kids[i] = -1;
    if (kids[i] == -1)
      avail++;
  }

  return(avail);
}


double
getTime(void) {
  struct timeval  tp;
  gettimeofday(&tp, NULL);
  return(tp.tv_sec + (double)tp.tv_usec / 1000000.0);
}


int
main(int argc, char **argv) {
  int   arg;
  int   verbose = 0;
  int   numCPUs = 4;
  char *scriptFileName = 0L;
  FILE *scriptFile;
  char *scriptLine = 0L;
  int   len;
  pid_t child, kids[1024];
  int   i;
  int   executedCount = 0;
  int   useShell = 0;
  int   waitTime = 10000;
  int   lineLength = 10241;
  double startTime = getTime() - 0.1;  //  Cheap way to avoid a divide-by-zero test later.

  arg = 1;
  while (arg < argc) {
    switch (argv[arg][1]) {
    case 'h':
      fprintf(stderr, "usage: %s [-v] [-s] [-c numCPUs] [-w us] [-l length] [-f scriptFile]\n", argv[0]);
      fprintf(stderr, "       -v  be verbose, print the number of jobs run and the number of jobs finished per second\n");
      fprintf(stderr, "       -s  use the shell (/bin/sh) to run the command\n");
      fprintf(stderr, "       -c  run numCPUs jobs concurrently (default = %d)\n", numCPUs);
      fprintf(stderr, "       -w  wait for us microseconds when busy (default = %d)\n", waitTime);
      fprintf(stderr, "       -l  lines are no longer than length characters (default = %d)\n", lineLength);
      fprintf(stderr, "       -f  use this script file, otherwise reads from stdin\n");
      exit(0);
      break;
    case 'v':
      verbose++;
      break;
    case 's':
      useShell = 1;
      break;
    case 'c':
      arg++;
      numCPUs = atoi(argv[arg]);
      break;
    case 'w':
      arg++;
      waitTime = atoi(argv[arg]);
      break;
    case 'l':
      arg++;
      lineLength = atoi(argv[arg]);
      break;
    case 'f':
      arg++;
      scriptFileName = argv[arg];
      break;
    default:
      fprintf(stderr, "Unknown option '%s'\n", argv[arg]);
      break;
    }
    arg++;
  }
    
  if (scriptFileName == 0L) {
    fprintf(stderr, "%s; scriptFile not given; using stdin.\n", argv[0]);
    scriptFile = stdin;
  } else {
    scriptFile = fopen(scriptFileName, "r");
  }

  scriptLine = (char *)malloc(sizeof(char) * lineLength);


  for (i=0; i<numCPUs; i++)
    kids[i] = -1;

  while (!feof(scriptFile)) {
    if (verbose) {
      fprintf(stderr, "Exec %d -- %f per second.\r", executedCount, executedCount / (getTime() - startTime));
      fflush(stderr);
      executedCount++;
    }

    while (reapChildren(kids, numCPUs) == 0)
      usleep(waitTime);

    fgets(scriptLine, lineLength, scriptFile);

    if (!feof(scriptFile)) {
      len = strlen(scriptLine);
      if (scriptLine[len-1] != '\n') {
        fprintf(stderr, "%s: WARNING: line doesn't end with a new line.  Possibly truncated.\n", argv[0]);
        fprintf(stderr, "%s: '%s'\n", argv[0], scriptLine);
      }
      
      child = runChild(useShell, scriptLine, len);

      //  Insert the child ID into the kids array at the first available spot
      //
      for (i=0; i<numCPUs; i++) {
        if (kids[i] == -1) {
          kids[i] = child;
          break;
        }
      }
    }
  }

  //  Wait (somewhat inefficiently) for kids to finish.
  //
  while (reapChildren(kids, numCPUs) != numCPUs)
    usleep(waitTime);

  return(0);
}
